.data <- rlang::.data

#' Generate rmarkdown codebook
#'
#' Pass a data frame to this function to make a codebook for that dataset.
#' If the dataset has metadata (attributes) set on its variables, these will be
#' used to make the codebook more informative. Examples are item, value, and
#' missing labels.
#' Data frames imported via [haven::read_dta()], [haven::read_sav()], or from
#' [formr.org](https://formr.org) will have these attributes in the right format.
#' By calling this function inside a knitr code chunk, the
#' codebook will become part of the document you are generating.
#'
#' @param results a data frame, ideally with attributes set on variables
#' @param reliabilities a named list with one entry per scale and one or several printable reliability computations for this scale. if NULL, computed on-the-fly using compute_reliabilities
#' @param survey_repetition defaults to "auto" which is to try to determine the level of repetition from the "session" and "created" variables. Other values are: single, repeated_once, repeated_many
#' @param survey_overview whether to print an overview of survey entries, durations (depends on presence of columns session, created, modified, ended, expired)
#' @param missingness_report whether to print a missingness report. Turn off if this gets too complicated and you need a custom solution (e.g. in case of random missings).
#' @param metadata_table whether to print a metadata table/tabular codebook.
#' @param metadata_json whether to include machine-readable metadata as JSON-LD (not visible)
#' @param indent add # to this to make the headings in the components lower-level. defaults to beginning at h2
#'
#' @export
#' @examples
#' # will generate figures in a temporary directory
#' old_base_dir <- knitr::opts_knit$get("base.dir")
#' knitr::opts_knit$set(base.dir = tempdir())
#' on.exit(knitr::opts_knit$set(base.dir = old_base_dir))
#' data("bfi")
#' bfi <- bfi[, c("BFIK_open_1", "BFIK_open_1")]
#' md <- codebook(bfi, survey_repetition = "single", metadata_table = FALSE)
codebook <- function(results, reliabilities = NULL,
    survey_repetition = c('auto', 'single', 'repeated_once', 'repeated_many'),
    survey_overview = TRUE,
    missingness_report = TRUE, metadata_table = TRUE,
    metadata_json = TRUE, indent = '#') {
  # todo: factor out the time stuff
  # todo: factor out the repetition detection stuff
  survey_repetition <- match.arg(survey_repetition)
  if (survey_repetition == "auto") {
    if (!(exists("session", results) &&
      exists("created", results) &&
      exists("ended", results))) {
      survey_repetition <- 'single'
      warning("The variables session, created, ended have to be defined for ",
              "automatic survey repetition detection to work. Set to no ",
              "repetition by default.")
    } else {
      users <- dplyr::n_distinct(results$session)
      rows_per_user <- nrow(results)/users

      dupes <- sum(duplicated(
        dplyr::select(results, .data$session, .data$created)))
      if (dupes > 0) {
        stop("There seem to be ", dupes, " duplicated rows in this survey ",
             "(duplicate session-created combinations)")
      }
      rows_by_user <- dplyr::count(dplyr::filter(results, !is.na(.data$ended)),
                                   .data$session)$n
      survey_repetition <- ifelse( rows_per_user <= 1,
                                  "single",
                                  ifelse(rows_by_user %in% 1:2,
                                         "repeated_once",
                                         "repeated_many"))
    }
  }

  if (is.null(reliabilities)) {
    reliabilities <- compute_reliabilities(results, survey_repetition)
  }

  df_name <- deparse(substitute(results))
  old_opt <- options('knitr.duplicate.label')$knitr.duplicate.label
  options(knitr.duplicate.label = 'allow')
  on.exit(options(knitr.duplicate.label = old_opt))
  options <- list(
    fig.path =
      paste0(knitr::opts_chunk$get("fig.path"), "cb_", df_name, "_"),
    cache.path =
      paste0(knitr::opts_chunk$get("cache.path"), "cb_", df_name, "_")
  )

  if (survey_overview &&
        exists("session", results) &&
        exists("created", results) &&
        exists("ended", results) &&
        exists("expired", results) &&
        exists("modified", results)) {
    survey_overview <- codebook_survey_overview(results, survey_repetition)
  } else {
    survey_overview <- ''
  }

  scales_items <- new.env()
  vars <- names(results)
  items_contained_in_scales <- c()
  `%<-%` <- future::`%<-%`
  for (i in seq_along(vars)) {
    var <- vars[i]
    scale <- results[[ var ]]
    scale_info <- attributes(scale)
    if ( !is.null(scale_info) && exists("scale_item_names", scale_info)) {
      items_contained_in_scales <- c(items_contained_in_scales,
                                     scale_info$scale_item_names, var)
      items <- dplyr::select(results,
                  rlang::UQS(rlang::quos(scale_info$scale_item_names)))
      scales_items[[var]] %<-% {tryCatch({
        codebook_component_scale(
          scale = scale, scale_name = var,
          items = items,
          reliabilities = reliabilities[[var]], indent = indent) },
      error = function(e) stop("Could not summarise scale ", var, ". ", e)) }
    }
  }

  dont_show_these <- c(items_contained_in_scales,
                       c("session", "created", "modified", "expired", "ended"))
  for (i in seq_along(vars)) {
    var <- vars[i]
    item <- results[[ var ]]
    if (var %in% dont_show_these) {
      next # don't do scales again
    } else {
      scales_items[[var]] %<-% {tryCatch({
                      codebook_component_single_item( item = item,
                              item_name = var, indent = indent ) },
      error = function(e) stop("Could not summarise item ", var, ". ", e)) }

    }
  }

  scales_items <- as.list(scales_items)
  # reorder
  done_vars <- intersect(vars, names(scales_items))
  scales_items <- scales_items[done_vars]

  if (missingness_report) {
    missingness_report <- codebook_missingness(results, indent = indent)
  } else {
    missingness_report <- ''
  }

  if (metadata_table) {
    items <- codebook_items(results, indent = indent)
  } else {
    items <- ''
  }
  if (metadata_json) {
    jsonld <- metadata_jsonld(results)
  } else {
    jsonld <- ''
  }

  asis_knit_child(require_file("_codebook.Rmd"), options = options)
}


#' Codebook survey overview
#'
#' An overview of the number of rows and groups, and of the durations participants
#' needed to respond (if those data are available).
#'
#' @param results a data frame which has the following columns: session, created, modified, expired, ended
#' @param survey_repetition defaults to single (other values: repeated_once, repeated_many). controls whether internal consistency, retest reliability or multilevel reliability is computed
#' @param indent add # to this to make the headings in the components lower-level. defaults to beginning at h2
#'
#' @export
#' @examples
#' # will generate figures in a figure/ subdirectory
#' data("bfi")
#' codebook_survey_overview(bfi)
codebook_survey_overview <- function(results, survey_repetition = "single",
                                     indent = "##") {
  stopifnot(exists("session", results))
  stopifnot(exists("created", results))
  stopifnot(exists("modified", results))
  stopifnot(exists("expired", results))
  stopifnot(exists("ended", results))

  users <- dplyr::n_distinct(results$session)
  finished_users <- dplyr::n_distinct(results[!is.na(results$ended),]$session)
  rows_per_user <- nrow(results)/users
  rows_by_user <- dplyr::count(dplyr::filter(results, !is.na(.data$ended)),
                               .data$session)$n

  duration <- dplyr::mutate(dplyr::filter(results, !is.na(ended)),
              duration = as.double(.data$ended - .data$created, unit = "mins"))
  upper_limit <- stats::median(duration$duration) +
                          4*stats::mad(duration$duration)
  high_vals <- sum(upper_limit < duration$duration)
  if (high_vals == 0) {
    upper_limit <- max(duration$duration)
  }
  lower_limit <- min(duration$duration)
  low_vals <- sum(lower_limit < 0)
  if (low_vals == 0) {
    lower_limit <- 0
  }

  started <- sum(!is.na(results$modified))
  only_viewed <- sum(is.na(results$ended) & is.na(results$modified))

  options <- list(
    fig.path = paste0(knitr::opts_chunk$get("fig.path"), "overview_"),
    cache.path = paste0(knitr::opts_chunk$get("cache.path"), "overview_")
  )
  asis_knit_child(require_file("_codebook_survey_overview.Rmd"),
                  options = options)
}


#' Codebook missingness
#'
#' An overview table of missingness patterns generated using [md_pattern()].
#'
#' @param results a data frame
#' @param indent add # to this to make the headings in the components lower-level. defaults to beginning at h2
#'
#' @export
#' @examples
#' data("bfi")
#' codebook_missingness(bfi)
codebook_missingness <- function(results, indent = "##") {
  options <- list(
    fig.path = paste0(knitr::opts_chunk$get("fig.path"), "overview_"),
    cache.path = paste0(knitr::opts_chunk$get("cache.path"), "overview_")
  )
  md_pattern <- md_pattern(results)
  asis_knit_child(require_file("_codebook_missingness.Rmd"), options = options)
}



#' Metadata as JSON-LD
#'
#' Echo a list of a metadata, generated using [metadata_list()] as JSON-LD in a
#' script tag.
#'
#' @param results a data frame, ideally with attributes set on variables
#' @export
#' @examples
#' data("bfi")
#' metadata_jsonld(bfi)
metadata_jsonld <- function(results) {
  options <- list(
    fig.path = paste0(knitr::opts_chunk$get("fig.path"), "metadata_"),
    cache.path = paste0(knitr::opts_chunk$get("cache.path"), "metadata_")
  )
  metadata <- metadata_list(results)
  asis_knit_child(require_file("_metadata_jsonld.Rmd"), options = options)
}


#' Tabular codebook
#'
#' Renders a tabular codebook including attributes and data summaries. The table
#' is generated using [DT::datatable()] and can be exported to CSV, Excel, etc.
#'
#'
#' @param results a data frame, ideally with attributes set on variables
#' @param indent add # to this to make the headings in the components lower-level. defaults to beginning at h2
#'
#' @export
#' @examples
#' data("bfi")
#' \dontrun{
#' # doesn't show interactively, because a html widget needs to be registered
#' codebook_items(bfi)
#' }
codebook_items <- function(results, indent = "##") {
  options <- list(
    fig.path = paste0(knitr::opts_chunk$get("fig.path"), "overview_"),
    cache.path = paste0(knitr::opts_chunk$get("cache.path"), "overview_")
  )
  metadata_table <- codebook_table(results)
  metadata_table <- dplyr::mutate(metadata_table,
         name = paste0('<a href="#', safe_name(.data$name), '">',
                       recursive_escape(.data$name), '</a>'))
  # bit ugly to suppress warnings here, but necessary for escaping whatever
  # columns there may be
  suppressWarnings(
    metadata_table <- dplyr::mutate_at(metadata_table, dplyr::vars(
    dplyr::one_of("label", "scale_item_names", "value_labels", "showif")),
    dplyr::funs(recursive_escape)) )

  asis_knit_child(require_file("_codebook_items.Rmd"), options = options)
}

#' Codebook component for scales
#'
#'
#' @param scale a scale with attributes set
#' @param scale_name the variable name of this scale
#' @param items a data.frame with the items constituting the scale
#' @param reliabilities a list with one or several results from calls to psych package functions for computing reliability
#' @param indent add # to this to make the headings in the components lower-level. defaults to beginning at h2
#'
#' @export
#' @examples
#' # will generate figures in a temporary directory
#' old_base_dir <- knitr::opts_knit$get("base.dir")
#' knitr::opts_knit$set(base.dir = tempdir())
#' on.exit(knitr::opts_knit$set(base.dir = old_base_dir))
#' data("bfi")
#' bfi <- bfi[,c("BFIK_open", paste0("BFIK_open_", 1:4))]
#' codebook_component_scale(bfi[,1], "BFIK_open", bfi[,-1],
#'    reliabilities = list(BFIK_open = psych::alpha(bfi[,-1])))
codebook_component_scale <- function(scale, scale_name, items, reliabilities,
                                     indent = '##') {
  stopifnot( exists("scale_item_names", attributes(scale)))
  stopifnot( attributes(scale)$scale_item_names %in% names(items) )
  safe_name <- safe_name(scale_name)

  options <- list(
    fig.path = paste0(knitr::opts_chunk$get("fig.path"), safe_name, "_"),
    cache.path = paste0(knitr::opts_chunk$get("cache.path"), safe_name, "_")
  )
  old_opt <- options('knitr.duplicate.label')$knitr.duplicate.label
  options(knitr.duplicate.label = 'allow')
  on.exit(options(knitr.duplicate.label = old_opt))

  asis_knit_child(require_file("_codebook_scale.Rmd"), options = options)
}

#' Codebook component for single items
#'
#'
#' @param item an item with attributes set
#' @param item_name the item name
#' @param indent add # to this to make the headings in the components lower-level. defaults to beginning at h2
#'
#' @export
#' @examples
#' # will generate figure in a temporary directory
#' old_base_dir <- knitr::opts_knit$get("base.dir")
#' knitr::opts_knit$set(base.dir = tempdir())
#' on.exit(knitr::opts_knit$set(base.dir = old_base_dir))
#' data("bfi")
#' codebook_component_single_item(bfi$BFIK_open_1, "BFIK_open_1")
codebook_component_single_item <- function(item, item_name, indent = '##') {
  safe_name <- safe_name(item_name)
  options <- list(
    fig.path = paste0(knitr::opts_chunk$get("fig.path"), safe_name, "_"),
    cache.path = paste0(knitr::opts_chunk$get("cache.path"), safe_name, "_")
  )
  asis_knit_child(require_file("_codebook_item.Rmd"), options = options)
}

#' Codebook metadata table
#'
#' will generate a table combining metadata from variable attributes
#' with data summaries generated using [skimr::skim_to_wide()]
#'
#' @param results a data frame, ideally with attributes set on variables
#'
#' @export
#' @examples
#' data("bfi")
#' codebook_table(bfi)
codebook_table <- function(results) {
  skimmed <- skim_to_wide_labelled(results)

  metadata <- dplyr::bind_rows(
    # var = results$session
  lapply(results, function(var) {
    x <- attributes(var)
    if (is.null(x)) {
      data.frame(label = NA)
    } else {
      if (exists("class", x)) {
        x$class <- NULL
      }
      if (exists("tzone", x)) {
        x$tzone <- NULL
      }
      if (exists("label", x)) {
        if (exists("item", x)) {
          if (exists("label", x$item)) {
            x$item$label <- NULL
          }
          if (exists("label_parsed", x$item)) {
            x$item$label_parsed <- NULL
          }
        }
      }
      if (exists("labels", x)) {
        if (!is.null(names(x$labels))) {
          x$value_labels <- paste(paste0(names(x$labels),
                                         "=", x$labels), collapse = ",")
        } else {
          x$value_labels <- paste(x$labels, collapse = ",")
        }
        x$labels <- NULL
        if (exists("item", x) && exists("choices", x$item)) {
          x$item$choices <- NULL
        }
      }

      if (exists("item", x) && exists("name", x$item)) {
        x$item$name <- NULL
      }
      if (exists("scale_item_names", x)) {
        x$scale_item_names <- paste(x$scale_item_names, collapse = ", ")
      }
      as.data.frame(t(purrr::flatten(x)))
    }
  }), .id = "name")

  metadata <- dplyr::left_join(metadata,
                               dplyr::rename(skimmed, data_type = .data$type),
                               by = c("name" = "variable"))
  order <- c("name", "label", "type", "type_options", "data_type", "ordered",
            "value_labels",  "optional", "showif",
            "scale_item_names",
            "value", "item_order", "block_order", "class",
            "missing", "complete", "n",  "empty", "n_unique",
            "top_counts", "count", "median", "min", "max",
            "mean", "sd", "p0", "p25", "p50", "p75", "p100", "hist")
  not_all_na <- function(x) { !all(is.na(x)) && !is.null(unlist(x)) }
  cols <- setdiff(union(
               intersect(order, names(metadata)), # order
              setdiff(names(metadata), order)), # include other cols
               c("choice_list"))

  metadata <- dplyr::select(metadata,  rlang::UQS(rlang::quos(cols)))
  dplyr::select_if(metadata, not_all_na )
}



#' Metadata from dataframe
#'
#' Returns a list containing variable metadata (attributes) and data summaries.
#'
#' @param results a data frame, ideally with attributes set on variables
#'
#' @export
#' @examples
#' data("bfi")
#' md_list <- metadata_list(bfi)
#' md_list$variableMeasured[[20]]
metadata_list <- function(results) {
  name <- deparse(substitute(results))
  metadata <- attributes(results)
  metadata$names <- NULL
  metadata$row.names <- NULL
  metadata$class <- NULL

  if (!exists("name", metadata)) {
    metadata$name <- name
  }

  if (!exists("keywords", metadata)) {
    metadata$keywords <- names(results)
  }

  metadata$variableMeasured <- lapply(names(results), function(var) {
      x <- attributes(results[[var]])
      x$name <- var

      if (is.null(x)) {
        x <- list()
      } else {
        if (exists("class", x)) {
          x$class <- NULL
        }
        if (exists("tzone", x)) {
          x$tzone <- NULL
        }
        if (exists("label", x)) {
          x$description <- x$label
          x$label <- NULL
        }
        if (exists("labels", x)) {
          x$value <- as.list(x$labels)
          # x$value[["@type"]] = "StructuredValue"
          x$maxValue <- max(x$labels, na.rm = TRUE)
          x$minValue <- min(x$labels, na.rm = TRUE)
          x$labels <- NULL
        }
        if (exists("item", x)) {
          if (exists("type", x$item)) {
            x$item$item_type <- x$item$type
            x$item$type <- NULL
          }
          if (exists("choices", x$item)) {
            x$item$choices[["@type"]] <- "http://formr.org/codebook/ItemChoices"
          }
          x$measurementTechnique <- "self-report"
          x$item[["@type"]] <- "http://formr.org/codebook/Item"
        }

      }
      x$data_summary <- skim_to_wide_labelled(results[[var]])
      x$data_summary$variable <- NULL
      if (exists("type", x$data_summary)) {
        if (!exists("value", x)) {
          x$value <- switch(x$data_summary$type,
              character = "text",
              integer = "Number",
              numeric = "Number",
              factor = "StructuredValue",
              labelled = "StructuredValue"
          )
        }
        x$data_summary$type <- NULL
      }
      x$data_summary[["@type"]] <- "http://formr.org/codebook/SummaryStatistics"
      x[["@context"]] <- list("@vocab" = "http://pending.schema.org/")
      x[["@type"]] <- "propertyValue"
      x
    })

  metadata[["@context"]] <- "http://schema.org/"
  metadata[["@type"]] <- "Dataset"

  metadata
}


skim_to_wide_labelled <- function(...) {
  suppressMessages(
    skimr::skim_with(labelled = skimr::get_skimmers()$factor,
                     labelled_spss = skimr::get_skimmers()$factor)
  )
  on.exit(skimr::skim_with_defaults())
  skimr::skim_to_wide(...)
}


